
/* EQUATION SPECIFIC FVM routines
 * Shallow Water equations
 */

///::title:: Shallow Water Equations
///::vars:: GRAVITY=9.812 BED_OFFSET=0.0 BED_SCALING=1.0 CFL=0.45
///::consts:: TOPOGRAPHY=1 EC_TOPOGRAPHY_QUAD=0 INTERPOLATE_TOPOGRAPHY=0
///::depends:: topography

#include "equation.h"
#include "ec_util.h"

void calc_point_flux(ConservedVars &F, const AllVariables &V, int d)
{
  if (d != 2) {
    F.h       = V.U.h * V.XU.u[d];
    F.hu[d]   = V.U.h * sqr(V.XU.u[d]) + GRAVITY * sqr(V.U.h) / 2.0;
    F.hu[1-d] = V.U.h * V.XU.u[d] * V.XU.u[1-d];
  }
}

void apply_constraints (ConservedVars &CV, XtraVars &XV)
{
  // Constraints for ConservedVars
  if (CV.h < 0.0) CV.h = 0.0;
  // Constraints for XtraVars
  if (XV.E < 0.0) XV.E = 0.0;
}

real calc_time_step()
{
  vector maxeig;
  maxeig.zero();
  real sqrt_gh;
  
  for (int i = 0; i < NXl; i++)
    for (int j = 0; j < NYl; j++) {
      sqrt_gh = sqrt ( GRAVITY * U[i][j][0].h );
      for (int d = 0; d < NDIR; d++) {
        
        if (!HAVE_DIR(d)) continue;
        
        // skips dry cells
        if (U[i][j][0].h == 0.0) continue;
        
        maxeig[d] = max( maxeig[d], fabs(Xtra[i][j][0].u[d]) + sqrt_gh );
      }
    }
  
  return CFL / ( maxeig[X]/dx[X] + maxeig[Y]/dx[Y] );
}

void calc_xtra(XtraVars &Xijk, const ConservedVars &Uijk, int i, int j, int k)
{
  Xijk.u = Uijk.hu / Uijk.h;
  Xijk.E = ( Uijk.h * (Xijk.u * Xijk.u) + GRAVITY * sqr(Uijk.h) ) / 2.0;
  Xijk.b = Xtra[i][j][k].b;
}

void calc_all_vars(AllVariables &V, const PrimitiveVars &P)
{
  P.toConservedVars(V.U);
  V.XU.u = P.u;
  V.XU.E = ( P.h * (P.u * P.u) + GRAVITY * sqr(P.h) ) / 2.0;
}

#if INDICATOR == 1
real indicator(int i, int j, int k, int d)
{
  ConservedVars &Uijk = U[i][j][k];
  return Uijk.h;
}
#endif

#if ENTROPIC_RECONSTRUCTION

void ConservedVars_to_ReconstVars (ConservedVars &CV, const XtraVars &XV) {
  // converts conserved variables to entropy variables
  CV.h += XV.b;
  CV.h *= GRAVITY;
  CV.h -= sqr(XV.u[X]) / 2.0;
  CV.h -= sqr(XV.u[Y]) / 2.0;
}

void ReconstVars_to_ConservedVars (ConservedVars &CV, const XtraVars &XV) {
  // converts entropy variables to conserved variables
  CV.h += sqr(XV.u[X]) / 2.0;
  CV.h += sqr(XV.u[Y]) / 2.0;
  CV.h /= GRAVITY;
  CV.h -= XV.b;
}

void ReconstVars_to_ScaledVars (ConservedVars &CV, const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d) {
  // converts entropy variables to scaled Roe variables
  // multiplication by R^T
  real RT [3][3];
  compute_R (RT, lCV, rCV, lXV, rXV, d);
  transpose (RT);
  
  real W [3];
  
  W[0] = CV.h;
  W[1] = CV.hu[X];
  W[2] = CV.hu[Y];
  
  zero(CV);
  for (int a=0; a<3; a++) {
    CV.h     += RT[0][a] * W[a];
    CV.hu[X] += RT[1][a] * W[a];
    CV.hu[Y] += RT[2][a] * W[a];
  }
}

void ScaledVars_to_ReconstVars (ConservedVars &CV, const ConservedVars &lCV, const ConservedVars &rCV, const XtraVars &lXV, const XtraVars &rXV, int d) {
  // converts scaled Roe variables to entropy variables
  // multiplication by the inverse of R^T
  real RTinv [3][3];
  compute_Rinv (RTinv, lCV, rCV, lXV, rXV, d);
  transpose (RTinv);
  
  real W [3];
  
  W[0] = CV.h;
  W[1] = CV.hu[X];
  W[2] = CV.hu[Y];
  
  zero(CV);
  for (int a=0; a<3; a++) {
    CV.h     += RTinv[0][a] * W[a];
    CV.hu[X] += RTinv[1][a] * W[a];
    CV.hu[Y] += RTinv[2][a] * W[a];
  }
}

#endif
