
#include <cstdlib>
#include <iostream>

#include "main.h"
#include "saver.h"
#include "rng.h"
#include "chaos.h"
#include "multi.h"

///::title:: Modelling software for Stochastic systems of Hyperbolic Partial Differential Equations
///::consts:: GHOST_SIZE=1 HAVE_GRAVITY=0 INDICATOR=0 ENTROPIC_RECONSTRUCTION=0
///::vars:: CFL=0.9 ORDER=1 NSAVES=1 SAVE0=0 WRITE_DATAFILES=1 PRINT_t=0
///::vars:: DATA_FILE="output.dat" DATA_PATH="" TIME_FILE="times.py" RUNTIME_FILE="runtime.py"
/////Defaults for models:
///::vars:: NX=1 NY=1 NZ=1
///::consts:: MAXX=1 MAXY=1 MAXZ=1
///::vars:: ML=1 L=0 DOWNSAMPLE=0 COMPUTE_VARIANCE=1 VARIANCE_FIX=1 COMPUTE_KURTOSIS=0
///::vars:: PRIME_SHIFT=10 RUN_NUMBER=1 SAMPLES_OFFSET=0 EXP_LOG=1
///::vars:: LOAD_BALANCING_TYPE=1
///::consts:: SKIP_IO_WTIME=1
///::consts:: N_UNIFORM=0 N_NORMAL=0 T_UNIFORM=0 T_NORMAL=0 K_UNIFORM=0 K_NORMAL=0
///::consts:: NO_POWELL=1
/////Defaults for hooks:
///::macro:: INIT_HOOK @ @
///::macro:: LEVEL_INIT_HOOK @ @
///::macro:: PRE_STEP_HOOK @ @
///::macro:: SOURCE_HOOK @ @
///::macro:: POST_STEP_HOOK @ @
///::macro:: FINAL_HOOK @ @
///::depends:: structures equation bc chaos

// === MLMC

int multi_domain_root, multi_samplers_root, multi_level_root;
int multi_rank, multi_rank_samplers;
int MULTI_TOTAL, LOWEST_MPI_LEVEL;
int MULTIXl, MULTIYl, MULTIZl;
int NXl, NYl, NZl;

// container for random structures
Chaos chaos;

// additional variables
long_int M[L+1];                 // number of samples for each level
int N[NDIR][L+1];                // number of cells for each level

int MULTI[NDIR][L+1];            // number node groups working on the same domain
int NSAMPLERS[L+1];              // number node groups working on the same level samples
int NSAMPLERSl;                  // number node groups working on the same level samples
int ACTIVE_SAMPLERSl;            // number node groups working on the same downsampled level samples
int level_begin, level_end;      // levels to start and to end (might be different from 0 and L for multi:mpi)
long_int partition_size;         // numbe of samples per one sampler

// === volumes for statistical information and performance optimization

volume <ConservedVars, 1> mean_U [NSAVES+SAVE0];
volume <XtraVars, 1>      mean_X [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
volume <ConservedVars, 1> var_U  [NSAVES+SAVE0];
volume <XtraVars, 1>      var_X  [NSAVES+SAVE0];
#endif

volume <ConservedVars, 1> mean_U_l [NSAVES+SAVE0];
volume <XtraVars, 1>      mean_X_l [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
volume <ConservedVars, 1> var_U_l  [NSAVES+SAVE0];
volume <XtraVars, 1>      var_X_l  [NSAVES+SAVE0];
#endif

#if DOWNSAMPLE
volume <ConservedVars, 1> mean_U_l_downsampled [NSAVES+SAVE0];
volume <XtraVars, 1>      mean_X_l_downsampled [NSAVES+SAVE0];
#if COMPUTE_VARIANCE
volume <ConservedVars, 1> var_U_l_downsampled  [NSAVES+SAVE0];
volume <XtraVars, 1>      var_X_l_downsampled  [NSAVES+SAVE0];
#endif
#endif

#if COMPUTE_VARIANCE
volume <ConservedVars, 1> temp_U_l;
volume <XtraVars,      1> temp_X_l;
#endif

// === MLMC routines

// computes MULTI[:][:], M[:], SAMPLERS[:], N[:][:], MULTI_TOTAL, LOWEST_MPI_LEVEL
void mlmc_init () {
  
  // !!! IF MODIFYING, also CHECK make.py::MakeMLMC::get_multi_total() for possible necessary modifications
  
  // === computes level resolutions and parallelisation parameters | N[:][:], MULTI[:][:], M[:], SAMPLERS[:]
  
  MULTI[X][L] = MULTIX;
  MULTI[Y][L] = MULTIY;
  MULTI[Z][L] = MULTIZ;
  
  M[L] = ML;
  
  NSAMPLERS[L] = MULTIM;
  
  N[X][L] = NX / MULTI[X][L];
  N[Y][L] = NY / MULTI[Y][L];
  N[Z][L] = NZ / MULTI[Z][L];
  
  for (int l=L-1; l>=0; l--) {
    
    if (FACTOR * M[l+1] <= M_MAX)
      M[l]  = FACTOR * M[l+1];
    else {
      std::cerr << " :: ERROR: M[l] exceeded maximum value of " << M_MAX << ". Change typedef of 'long_int' in 'numtypes.h'." << std::endl;
      throw 1;
    }
    
    // maximal cummulative factor allowed in: MULTI[.][l] = MULTI[.][l+1] / 2. Ensures that: NSAMPLERS[l] <= M[l+1] for all l in {L-1, ..., 0}.
    int MULTI_FACTOR_MAX = WORK_FACTOR * M[l+1] / NSAMPLERS[l+1];
    
    // if DOWNSAMPLE=1, then no recomputation is present; hence NSAMPLERS[l+1] is no longer of high importance for level = l.
    for (int d=0; d<NDIR; d++) {
      if ( MULTI[d][l+1] != 1 && (MULTI_FACTOR_MAX > 1 || DOWNSAMPLE) ) {
        MULTI_FACTOR_MAX /= 2;
        MULTI[d][l] = MULTI[d][l+1] / 2;
      } else {
        MULTI[d][l] = MULTI[d][l+1];
      }
    }
    
    int MULTI_FACTOR  = (MULTI[X][l+1] / MULTI[X][l]) * (MULTI[Y][l+1] / MULTI[Y][l]) * (MULTI[Z][l+1] / MULTI[Z][l]);
    
    if ( (NSAMPLERS[l+1] * MULTI_FACTOR) >= WORK_FACTOR )
      NSAMPLERS[l] = (NSAMPLERS[l+1] * MULTI_FACTOR) / WORK_FACTOR;
    else
      NSAMPLERS[l] = 1;
    
    N[X][l] = N[X][l+1] / 2 * (MULTI[X][l+1] / MULTI[X][l]);
    
    if ( NY > 1 )
      N[Y][l] = N[Y][l+1] / 2 * (MULTI[Y][l+1] / MULTI[Y][l]);
    else
      N[Y][l] = N[Y][l+1];
    
    if ( NZ > 1 )
      N[Z][l] = N[Z][l+1] / 2 * (MULTI[Z][l+1] / MULTI[Z][l]);
    else
      N[Z][l] = N[Z][l+1];
    
    if (N[X][l] == 0 || N[Y][l] == 0 || N[Z][l] == 0) {
      std::cerr << " :: ERROR: Too many levels, no further coarsening possible" << std::endl;
      throw 1;
    }
  }
  
  // === computes how many levels should be parallelised | LOWEST_MPI_LEVEL
  
  LOWEST_MPI_LEVEL = 0;
  
  // if all levels do the same amount of work, all of them should be parallelised
  if (WORK_FACTOR != 1) {
    // otherwise find the level from which on parallelisation is no longer needed
    for (int l=L-1; l>=0; l--) {
      if (NSAMPLERS[l] == 1 && NSAMPLERS[l+1] == 1
          && MULTI[X][l]   == 1 && MULTI[Y][l]   == 1 && MULTI[Z][l]   == 1
          && MULTI[X][l+1] == 1 && MULTI[Y][l+1] == 1 && MULTI[Z][l+1] == 1)
      {
        LOWEST_MPI_LEVEL = l;
        break;
      }
    }
  }
  
  // === computes needed number of nodes | MULTI_TOTAL
  
  MULTI_TOTAL = 0;
  for (int l=L; l>=LOWEST_MPI_LEVEL; l--) {
    MULTI_TOTAL += NSAMPLERS[l] * MULTI[X][l] * MULTI[Y][l] * MULTI[Z][l];
  }
}

void volumes_init () {
  
  int l = level_end;
  
  for(int s = 0; s<NSAVES+SAVE0; s++) {
    
    // resizing array to match size of the finest level computed on this particular node
    mean_U [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    mean_X [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    #if COMPUTE_VARIANCE
    var_U  [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    var_X  [s] .reinit (N[X][l], N[Y][l], N[Z][l]);
    #endif
    
    // for stabilization of ML-part with exp-log
    #if EXP_LOG
      mean_U [s] .exp();
      mean_X [s] .exp();
      #if COMPUTE_VARIANCE
      var_U  [s] .exp();
      var_X  [s] .exp();
      #endif
    #endif
  }
}

// setup cores for domain decomposition and samplers
void multi_level_init (int l) {
  MULTIXl = MULTI[X][l];
  MULTIYl = MULTI[Y][l];
  MULTIZl = MULTI[Z][l];
  
  NSAMPLERSl = NSAMPLERS[l];
}
  
// setup meshes for level l
void level_init (int l) {
  NXl = N[X][l];
  NYl = N[Y][l];
  NZl = N[Z][l];
  
  // reinit volumes for resolutions | NXl, NYl, NZl
  U      .reinit (NXl, NYl, NZl);
  Xtra   .reinit (NXl, NYl, NZl);
  Source .reinit (NXl, NYl, NZl);
  for (int i=0; i<NDIR; i++)
    Flux[i].reinit (NXl, NYl, NZl);
}

unsigned int getSeed (unsigned int level) {
  
  unsigned int pos = PRIME_SHIFT + (RUN_NUMBER-1)*L + level;
  
  if (pos >= NPRIMES) {
    if (multi_main) {
      std::cerr << " :: ERROR: Prime number array 'primes' in primes.h is too small!" << std::endl;
      std::cerr << "  -> Decrease PRIME_SHIFT or L or RUN_NUMBER or increace the 'primes' array." << std::endl;
    }
    multi_finalise();
    std::exit(1);
  }
  
  return primes[pos];
}

// samples are generated in _each_ node with an appropriate offset
void rng_and_chaos_init (int l, int level ) {
  rng_init ( getSeed(level) );
  chaos_init ( SAMPLES_OFFSET + multi_rank_samplers * (M[l]/NSAMPLERSl) );
}

void init_sample () {
  t = 0;
  model_init();
  bc_init();
  step_finalize();      // "finalizes" the trivial timestep t=0
}

void run_sample () {
  while (t < t_halt) {
    time_step();
    step_finalize();
    POST_STEP_HOOK
  }
}

void run_samples (int l, char action, bool compute) {
  
  // initializing level l volumes for statistical information
  for(int s = 0; s<NSAVES+SAVE0; s++) {
    mean_U_l [s] .reinit (NXl, NYl, NZl);
    mean_X_l [s] .reinit (NXl, NYl, NZl);
    #if COMPUTE_VARIANCE
    var_U_l  [s] .reinit (NXl, NYl, NZl);
    var_X_l  [s] .reinit (NXl, NYl, NZl);
    #endif
  }
  
  #if COMPUTE_VARIANCE
  // partition size for MPI calls
  partition_size = M[l]/NSAMPLERSl;
  
  // temporary volumes for computational efficiency
  temp_U_l .reinit (NXl, NYl, NZl);
  temp_X_l .reinit (NXl, NYl, NZl);
  #endif
  
  #if COMPUTE_KURTOSIS
  volume <ConservedVars, 1> M2 [NSAVES+SAVE0];
  volume <ConservedVars, 1> M3 [NSAVES+SAVE0];
  volume <ConservedVars, 1> M4 [NSAVES+SAVE0];
  
  for (int s=0; s<NSAVES+SAVE0; s++) {
    M2[s] .reinit (NXl, NYl, NZl);
    M3[s] .reinit (NXl, NYl, NZl);
    M4[s] .reinit (NXl, NYl, NZl);
  }
  
  volume <ConservedVars, 1> delta (NXl, NYl, NZl);
  volume <ConservedVars, 1> delta_n (NXl, NYl, NZl);
  volume <ConservedVars, 1> delta_n2 (NXl, NYl, NZl);
  volume <ConservedVars, 1> term1 (NXl, NYl, NZl);
  #endif
  
  // compute samples
  if (compute) {
    
    #if DOWNSAMPLE
    if (l != L) {
      
      for (int s=0; s<NSAVES+SAVE0; s++) {
        mean_U_l_downsampled [s] .reinit (NXl, NYl, NZl);
        mean_X_l_downsampled [s] .reinit (NXl, NYl, NZl);
        #if COMPUTE_VARIANCE
        var_U_l_downsampled  [s] .reinit (NXl, NYl, NZl);
        var_X_l_downsampled  [s] .reinit (NXl, NYl, NZl);
        #endif
      }
      
      // samplers that work on 'downsampled' set of samples
      ACTIVE_SAMPLERSl = M[l+1] / (M[l]/NSAMPLERS[l]);
    }
    #endif
    
    for (long_int i=1; i<=M[l]/NSAMPLERSl; i++) {
      
      // draw a sample
      chaos = chaos_draw();
      
      // initialize initial data and bc
      init_sample();
      
      // run the simulation while saving the snapshots
      for (int s=0; s<NSAVES+SAVE0; s++) {
        
        // define when simulation should halt
        if (SAVE0)
          t_halt = 1.0 * MAXT / NSAVES * (s);
        else
          t_halt = 1.0 * MAXT / NSAVES * (s+1);
        
        // run simulation until it halts
        run_sample();
        
        // record the halt time
        if (i == 1) halt_times[s] = t;
        
        // compute the statistics for halt time
        
        // ConservedVars
        
        #if COMPUTE_VARIANCE
          
          #if COMPUTE_KURTOSIS
          delta.zero();
          delta_n.zero();
          delta_n2.zero();
          term1.zero();
          
          delta += U;
          delta -= mean_U_l[s];
          
          delta_n += delta;
          delta_n /= i;
          
          delta_n2 += delta_n;
          delta_n2 *= delta_n;
          
          term1 += delta;
          term1 *= delta_n;
          term1 *= i-1;
          
          mean_U_l[s] += delta_n;
          
          M4[s] += term1 * delta_n2 * (i*i - 3*i + 3);
          M4[s] += delta_n2 * M2[s] * 6;
          M4[s] -= delta_n * M3[s] * 4;
          
          M3[s] += term1 * delta_n * (i-2);
          M3[s] -= delta_n * M2[s] * 3;
          
          M2[s] += term1;
          
          #else
          
          if (i>1) {
            temp_U_l.zero();
            temp_U_l += U;
            temp_U_l -= mean_U_l[s]/(i-1);
            temp_U_l.sqr();
            temp_U_l *= (1 - 1.0/i);
            var_U_l[s] += temp_U_l;
          }
          
          #endif
          
        #endif
        
        #if !COMPUTE_KURTOSIS
        mean_U_l[s] += U;
        #endif
        
        // XtraVars
        
        #if COMPUTE_VARIANCE
        if (i>1) {
          temp_X_l.zero();
          temp_X_l += Xtra;
          temp_X_l -= mean_X_l[s]/(i-1);
          temp_X_l.sqr();
          temp_X_l *= (1 - 1.0/i);
          var_X_l[s] += temp_X_l;
        }
        #endif
        mean_X_l[s] += Xtra;
        
        #if DOWNSAMPLE
        if (l != L) {
          
          // all needed samples are in master sampler
          if (ACTIVE_SAMPLERSl <=1 && multi_samplers_root && i == M[l+1]) {
            mean_U_l_downsampled [s] += mean_U_l [s];
            mean_X_l_downsampled [s] += mean_X_l [s];
            #if COMPUTE_VARIANCE
            var_U_l_downsampled  [s] += var_U_l  [s];
            var_X_l_downsampled  [s] += var_X_l  [s];
            #endif
          }
          
          // all needed samples are in first ACTIVE_SAMPLERSl samplers
          else if (multi_rank_samplers < ACTIVE_SAMPLERSl && i == M[l]/NSAMPLERSl) {
            mean_U_l_downsampled [s] += mean_U_l [s];
            mean_X_l_downsampled [s] += mean_X_l [s];
            #if COMPUTE_VARIANCE
            var_U_l_downsampled  [s] += var_U_l  [s];
            var_X_l_downsampled  [s] += var_X_l  [s];
            #endif
          }
        }
        #endif
      }
    }
    
    #if COMPUTE_KURTOSIS
    for (int s=0; s<NSAVES+SAVE0; s++) {
      var_U_l [s] += M4[s];
      var_U_l [s] /= M[l] - 1;
    }
    #endif
    
    // all samplers took part
    ACTIVE_SAMPLERSl = NSAMPLERSl;
  }
  
  // reuse computed samples
  #if DOWNSAMPLE
  else {
    // samplers that worked on 'downsampled' set of samples
    ACTIVE_SAMPLERSl = M[l] / (M[l-1]/NSAMPLERS[l-1]);
    
    #if COMPUTE_VARIANCE
    // partition size of the previous level
    partition_size = M[l-1]/NSAMPLERSl;
    #endif
    
    if ( multi_rank_samplers < max (1, ACTIVE_SAMPLERSl) ) {
      for (int s=0; s<NSAVES+SAVE0; s++) {
        mean_U_l [s] += mean_U_l_downsampled [s];
        mean_X_l [s] += mean_X_l_downsampled [s];
        #if COMPUTE_VARIANCE
        var_U_l  [s] += var_U_l_downsampled  [s];
        var_X_l  [s] += var_X_l_downsampled  [s];
        #endif
      }
    }
  }
  #endif
  
  multi_combine_samples();
  
  if (multi_samplers_root) {
    for (int s=0; s<NSAVES+SAVE0; s++) {
      
      #if !COMPUTE_KURTOSIS
      
      // computing mean & variance
      mean_U_l [s] /= M[l];
      mean_X_l [s] /= M[l];
      
      #if COMPUTE_VARIANCE
      var_U_l [s] /= M[l] - 1;
      var_X_l [s] /= M[l] - 1;
      #endif
      
      #endif
      
      // for stabilization of ML-part with exp-log
      #if EXP_LOG
      
      mean_U_l [s] .exp();
      mean_X_l [s] .exp();
      #if COMPUTE_VARIANCE
      var_U_l  [s] .exp();
      var_X_l  [s] .exp();
      #endif
    
      // multiplying/dividing the final mean & variance to/by the details
      switch (action) {
        case '+' :
          mean_U [s] *= mean_U_l[s];
          mean_X [s] *= mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] *= var_U_l[s];
          var_X [s] *= var_X_l[s];
          #endif
          break;
        
        case '-' :
          mean_U [s] /= mean_U_l[s];
          mean_X [s] /= mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] /= var_U_l[s];
          var_X [s] /= var_X_l[s];
          #endif
      }
      
      // EXP_LOG=0
      #else
      // adding/subtracting details to/from the final mean & variance
      switch (action) {
        case '+' :
          mean_U [s] += mean_U_l[s];
          mean_X [s] += mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] += var_U_l[s];
          var_X [s] += var_X_l[s];
          #endif
          break;
        
        case '-' :
          mean_U [s] -= mean_U_l[s];
          mean_X [s] -= mean_X_l[s];
          
          #if COMPUTE_VARIANCE
          var_U [s] -= var_U_l[s];
          var_X [s] -= var_X_l[s];
          #endif
      }
      
      // EXP_LOG
      #endif
    }
  }
}

// === main routine

int main(int argc, char **argv) {
  
  mlmc_init();
  
  multi_init(argc, argv);
  INIT_HOOK
  
  // check for condition s <= (d+1)/2
  if (ORDER > ACTIVE_DIRS + 1 && multi_main) {
    std::cout << " :: ERROR: 2s <= d+1 is violated: " << ORDER << " > " << ACTIVE_DIRS + 1 << std::endl;
    multi_finalise();
    std::exit(1);
  }
  
  // check for condition MULTIM <= ML
  if (MULTIM > ML && multi_main) {
    std::cout << " :: ERROR: MULTIM <= ML is violated: " << MULTIM << " > " << ML << std::endl;
    multi_finalise();
    std::exit(1);
  }
  
  // DOWNSAMPLE=1 and LOAD_BALANCING_TYPE=2 are incompatible!
  if (DOWNSAMPLE == 1 && LOAD_BALANCING_TYPE == 2  && multi_main) {
    std::cout << " :: ERROR: DOWNSAMPLE=1 and LOAD_BALANCING_TYPE=2 are incompatible!" << std::endl;
    multi_finalise();
    std::exit(1);
  }
  
  volumes_init();
  
  #ifndef QUIET
  if (multi_main) std::cout << " :: INFO: total number of samples:  " << M_TOTAL << std::endl;
  if (multi_main) std::cout << " :: INFO: number of samples at L=0: " << M[0] << std::endl;
  if (multi_main) std::cout << " :: INFO: coarsest mesh level L0:   " << L0 << std::endl;
  #endif
  
  start_timer();
  
  // compute MC estimates for levels l, where: level_begin <= l <= level_end
  for (int l=level_begin; l<=level_end; l++) {
    
    // initialize to the mesh level l
    level_init(l);
    multi_level_init(l);
    multi_level_setup();
    LEVEL_INIT_HOOK
    
    #ifndef QUIET
    if (multi_level_root) {
      std::cout << " :: LEVEL " << l << ": resolution: " << NXl*MULTIXl << "x" << NYl*MULTIYl << "x" << NZl*MULTIZl;
      std::cout << ", samples: " << M[l];
      std::cout << ", cores: (" << MULTIXl << "*" << MULTIYl << "*" << MULTIZl << ")*" << NSAMPLERSl;
      std::cout << "=" << MULTIXl*MULTIYl*MULTIZl*NSAMPLERSl << "...";
    }
    #endif
    
    if (DOWNSAMPLE) {
      // generate RNG and chaos, use seed from level 0
      rng_and_chaos_init (l, 0);
    }
    else {
      // generate RNG and chaos, use seed from level l
      rng_and_chaos_init (l, l);
    }
    
    // get MC estimate of the first term for the current level
    run_samples (l, '+', true);
    
    #if LOAD_BALANCING_TYPE == 1
    
    // if level is not the last one, get MC estimate for the E_{M_{l+1}}[u_l]
    if ( l != L ) {
      
      if ( DOWNSAMPLE ) {
        // reuse existing results (no additional computing!) to get MC estimate of the second term for the upper level
        run_samples (l+1, '-', false);
      }
      else {
        // generate RNG and chaos for the upper level (use seed from level l+1)
        rng_and_chaos_init (l+1, l+1);
        
        // get MC estimate of the second term for the upper level
        run_samples (l+1, '-', true);
      }
    }
    
    #endif
    
    #if LOAD_BALANCING_TYPE == 2
    
    // if level is not 0, get MC estimate for the E_{M_l}[u_{l-1}]
    if ( l != 0 ) {
      // initialize to the mesh level l-1
      level_init(l-1);
      multi_level_setup();
      LEVEL_INIT_HOOK
      
      if (DOWNSAMPLE) {
        // generate RNG and chaos, use seed from level 0
        rng_and_chaos_init (l, 0);
      }
      else {
        // generate RNG and chaos, use seed from level l
        rng_and_chaos_init (l, l);
      }
      
      // get MC estimate of the second term
      run_samples (l, '-', true);
      
      // restore level setup
      if ( l == level_end )
        level_init(l);
    }
    
    #endif
    
    #ifndef QUIET
    if (multi_level_root) {
      std::cout << " done." << std::endl;
    }
    #endif
  }
  
  if (SKIP_IO_WTIME)
    stop_timer();
  
  #if L != 0
  #if WRITE_DATAFILES
  multi_combine_levels();
  #endif
  #endif
  
  multi_cleanup();
  
  if (!SKIP_IO_WTIME)
    stop_timer();
  
  #if WRITE_DATAFILES
  // saving data
  if ( multi_samplers_root && level_end == L ) {
    
    #ifndef QUIET
    if (multi_main) std::cout << " :: Writing datafiles...";
    #endif
    
    // saving mean
    Saver saver_mean("mean");
    saver_mean.snapshots();
    
    // saving variance (if computed)
    #if COMPUTE_VARIANCE
    Saver saver_var("variance");
    saver_var.snapshots();
    #endif
    
    #ifndef QUIET
    if (multi_main) std::cout << " done." << std::endl;
    #endif
  }
  #endif
  
  FINAL_HOOK
  multi_finalise();
}
